/* -*-c-*-
 *
 * Miscellaneous utilities (not Catacomb-specific)
 *
 * (c) 2005 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of the Python interface to Catacomb.
 *
 * Catacomb/Python is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Catacomb/Python is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Catacomb/Python; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "catacomb-python.h"

/* #undef HAVE_LONG_LONG */

/*----- External values ---------------------------------------------------*/

PyObject *modname = 0;
PyObject *home_module = 0;

/*----- Conversions -------------------------------------------------------*/

PyObject *getulong(unsigned long w)
{
  if (w <= LONG_MAX)
    return (PyInt_FromLong(w));
  else
    return (PyLong_FromUnsignedLong(w));
}

#ifndef HAVE_LONG_LONG
static PyObject *i32 = 0;
static int init_i32(void)
  { if (!i32 && (i32 = PyInt_FromLong(32)) == 0) return (-1); return (0); }
#endif

PyObject *getk64(kludge64 u)
{
#ifdef HAVE_LONG_LONG
  return (PyLong_FromUnsignedLongLong(GET64(unsigned PY_LONG_LONG, u)));
#else
  PyObject *i = 0, *j = 0, *t;
  PyObject *rc = 0;

  if (init_i32()) goto end;
  if ((i = PyLong_FromUnsignedLong(HI64(u))) == 0) goto end;
  if ((t = PyNumber_InPlaceLshift(i, i32)) == 0) goto end;
  Py_DECREF(i); i = t;
  if ((j = PyLong_FromUnsignedLong(LO64(u))) == 0) goto end;
  if ((t = PyNumber_InPlaceOr(i, j)) == 0) goto end;
  Py_DECREF(i); i = t;
  if ((rc = PyNumber_Int(i)) == 0) goto end;
end:
  Py_XDECREF(i);
  Py_XDECREF(j);
  return (rc);
#endif
}

PyObject *getbool(int b)
{
  if (b) RETURN_TRUE;
  else RETURN_FALSE;
}

PyObject *abstract_pynew(PyTypeObject *ty, PyObject *arg, PyObject *kw)
{
  PyErr_SetString(PyExc_TypeError, "can't instantiate this class");
  return (0);
}

int convulong(PyObject *o, void *pp)
{
  long i;
  unsigned long *p = pp;
  PyObject *t;

  if (!o) VALERR("can't delete");
  if (PyInt_Check(o)) {
    i = PyInt_AS_LONG(o);
    if (i < 0) VALERR("must be nonnegative");
    *p = i;
  } else {
    if ((t = PyNumber_Long(o)) == 0) goto end;
    *p = PyLong_AsUnsignedLong(t);
    Py_DECREF(t);
    if (PyErr_Occurred()) goto end;
  }
  return (1);
end:
  return (0);
}

#ifdef HAVE_UINT64
#  define CONVu64(n) do {						\
     kludge64 k;							\
     uint64 t;								\
     if (!convk64(o, &k)) goto end;					\
     t = GET64(uint64, k);						\
     if (t > MASK##n) VALERR("out of range");				\
     *p = t;								\
   } while (0)
#else
#  define CONVu64(n) assert(!"shouldn't be possible")
#endif

#define CONVU_(n)							\
  int convu##n(PyObject *o, void *pp)					\
  {									\
    unsigned long u;							\
    uint##n *p = pp;							\
									\
    if (MASK##n > ULONG_MAX)						\
      CONVu64(n);							\
    else {								\
      if (!convulong(o, &u)) goto end;					\
      if (u > MASK##n) VALERR("out of range");				\
      *p = u;								\
    }									\
    return (1);								\
  end:									\
    return (0);								\
  }
DOUINTSZ(CONVU_)

int convuint(PyObject *o, void *pp)
{
  unsigned long u;
  unsigned *p = pp;

  if (!convulong(o, &u)) goto end;
  if (u > UINT_MAX) VALERR("out of range");
  *p = u;
  return (1);
end:
  return (0);
}

int convk64(PyObject *o, void *pp)
{
  PyObject *i = 0;
  int rc = 0;
#if HAVE_LONG_LONG
  unsigned PY_LONG_LONG t;
#else
  PyObject *t;
  uint32 lo, hi;
#endif

  if (!o) VALERR("can't delete");
#if HAVE_LONG_LONG
  if ((i = PyNumber_Long(o)) == 0) goto end;
  t = PyLong_AsUnsignedLongLong(i);
  if (t == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred()) goto end;
  ASSIGN64(*(kludge64 *)pp, t);
#else
  if (init_i32()) goto end;
  if ((i = PyNumber_Int(o)) == 0) goto end;
  lo = PyInt_AsUnsignedLongMask(i);
  if ((t = PyNumber_InPlaceRshift(i, i32)) == 0) goto end;
  Py_DECREF(i); i = t;
  hi = PyInt_AsUnsignedLongMask(i);
  if ((t = PyNumber_InPlaceRshift(i, i32)) == 0) goto end;
  Py_DECREF(i); i = t;
  if (PyObject_IsTrue(i)) VALERR("out of range");
  SET64(*(kludge64 *)pp, hi, lo);
#endif
  rc = 1;

end:
  Py_XDECREF(i);
  return (rc);
}

int convmpw(PyObject *o, void *pp)
{
  unsigned long u;
  unsigned *p = pp;

  if (!convulong(o, &u)) goto end;
  if (u > MPW_MAX) VALERR("out of range");
  *p = u;
  return (1);
end:
  return (0);
}

int convszt(PyObject *o, void *pp)
{
  unsigned long u;
  size_t *p = pp;

  if (!convulong(o, &u)) goto end;
  if (u > ~(size_t)0) VALERR("out of range");
  *p = u;
  return (1);
end:
  return (0);
}

int convbool(PyObject *o, void *pp)
{
  if (!o) VALERR("can't delete");
  *(int *)pp = PyObject_IsTrue(o);
  return (1);
end:
  return (0);
}

/*----- Type messing ------------------------------------------------------*/

static const PyTypeObject emptytype = { 0 };

void *newtype(PyTypeObject *metaty,
	      const PyTypeObject *skel,
	      const char *name)
{
  PyHeapTypeObject *ty =
    (PyHeapTypeObject *)_PyObject_GC_Malloc(_PyObject_VAR_SIZE(metaty, 0));
  if (!skel) skel = &emptytype;
  memcpy(ty, skel, sizeof(*skel));
  if (ty->ht_type.tp_base) Py_INCREF(ty->ht_type.tp_base);
#define COPY(blah) do {							\
    if (ty->ht_type.tp_as_##blah) {					\
      memcpy(&ty->as_##blah,						\
	     ty->ht_type.tp_as_##blah,					\
	     sizeof(ty->as_##blah));					\
      ty->ht_type.tp_as_##blah = &ty->as_##blah;			\
    }									\
  } while (0)
  COPY(number);
  COPY(sequence);
  COPY(mapping);
  COPY(buffer);
#undef COPY
  if (name)
    ty->ht_name = PyString_FromString(name);
  else if (ty->ht_type.tp_name)
    ty->ht_name = PyString_FromString(ty->ht_type.tp_name);
  if (ty->ht_name)
    ty->ht_type.tp_name = PyString_AS_STRING(ty->ht_name);
  DISCARD(PyObject_INIT(&ty->ht_type, metaty));
  Py_INCREF(metaty);
  return (ty);
}

void typeready(PyTypeObject *ty)
{
  PyType_Ready(ty);
  PyDict_SetItemString(ty->tp_dict, "__module__", modname);
}

PyTypeObject *inittype(PyTypeObject *tyskel, PyTypeObject *meta)
{
  PyTypeObject *ty = newtype(meta, tyskel, 0);
  ty->tp_flags |= Py_TPFLAGS_HEAPTYPE;
  typeready(ty);
  return (ty);
}

/*----- Constants ---------------------------------------------------------*/

void setconstants(PyObject *mod, const struct nameval *c)
{
  PyObject *x;
  unsigned long u;

  while (c->name) {
    u = c->value;
    if (u <= LONG_MAX) x = PyInt_FromLong(u);
    else if (c->f&CF_SIGNED) x = PyInt_FromLong(-1 - (long)(ULONG_MAX - u));
    else x = PyLong_FromUnsignedLong(u);
    PyModule_AddObject(mod, (/*unconst*/ char *)c->name, x); c++;
  }
}

/*----- Building method tables --------------------------------------------*/

static PyMethodDef *global_methods;
static size_t nmethods, methodsz;

void addmethods(const PyMethodDef *m)
{
  size_t n, want, newsz;

  for (n = 0; m[n].ml_name; n++);
  want = nmethods + n + 1;
  if (want > methodsz) {
    newsz = methodsz ? 2*methodsz : 16;
    while (want > newsz) newsz *= 2;
    if (!global_methods)
      global_methods = PyObject_Malloc(newsz*sizeof(PyMethodDef));
    else
      global_methods = PyObject_Realloc(global_methods,
					newsz*sizeof(PyMethodDef));
    assert(global_methods);
    methodsz = newsz;
  }
  memcpy(global_methods + nmethods, m, n*sizeof(PyMethodDef));
  nmethods += n;
  global_methods[nmethods].ml_name = 0;
}

PyMethodDef *donemethods(void) { return (global_methods); }

/*----- Exceptions --------------------------------------------------------*/

PyObject *mkexc(PyObject *mod, PyObject *base,
		const char *name, PyMethodDef *mm)
{
  PyObject *nameobj = 0;
  PyObject *dict = 0;
  PyObject *exc = 0;
  PyObject *func = 0;
  PyObject *meth = 0;

  if ((dict = PyDict_New()) == 0) goto fail;

  if (mm) {
    while (mm->ml_name) {
      if ((func = PyCFunction_NewEx(mm, 0, mod)) == 0 ||
	  (meth = PyMethod_New(func, 0, exc)) == 0 ||
	  PyDict_SetItemString(dict, mm->ml_name, meth))
	goto fail;
      Py_DECREF(func); func = 0;
      Py_DECREF(meth); meth = 0;
      mm++;
    }
  }

  if ((nameobj = PyString_FromFormat("%s.%s",
				     PyModule_GetName(mod),
				     name)) == 0 ||
      (exc = PyErr_NewException(PyString_AS_STRING(nameobj),
				base, dict)) == 0)
    goto fail;

done:
  Py_XDECREF(nameobj);
  Py_XDECREF(dict);
  return (exc);

fail:
  Py_XDECREF(exc);
  Py_XDECREF(func);
  Py_XDECREF(meth);
  exc = 0;
  goto done;
}

void report_lost_exception_v(struct excinfo *exc,
			     const char *why, va_list ap)
{
  PyObject *hookfn = 0;
  PyObject *whyobj = 0;
  PyObject *obj = 0;

  /* Make sure we start out without a pending exception, or this will get
   * really confusing.
   */
  assert(!PyErr_Occurred());

  /* Format the explanation. */
  if (why) whyobj = PyString_FromFormatV(why, ap);
  else { whyobj = Py_None; Py_INCREF(whyobj); }

  /* Find our home module's `lostexchook' function.  This won't work if
   * there's no module, or the function isn't defined, or it's `None'.
   */
  if (!home_module) goto sys;
  hookfn = PyObject_GetAttrString(home_module, "lostexchook");
  if (hookfn == Py_None) goto sys;
  else if (hookfn) ;
  else if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto ouch;
  else { PyErr_Clear(); goto sys; }

  /* Call the hook function. */
  obj = PyObject_CallFunction(hookfn, "(OOOO)",
			      whyobj, exc->ty, exc->val, exc->tb);
  if (!obj) goto ouch;
  goto end;

  /* Something went wrong reporting the problem. */
ouch:
  PySys_WriteStderr("\n!!! FAILURE REPORTING LOST EXCEPTION\n");
  PyErr_Print();
  /* drop through... */

  /* There was no hook, so try to do something sensible using
   * `sys.excepthook'.
   */
sys:
  PySys_WriteStderr("\n!!! LOST EXCEPTION: %s\n",
		    PyString_AS_STRING(whyobj));
  RESTORE_EXCINFO(exc);
  PyErr_Print();
  /* drop through... */

  /* Clean up afterwards. */
end:
  Py_XDECREF(hookfn);
  Py_XDECREF(whyobj);
  Py_XDECREF(obj);
}

void report_lost_exception(struct excinfo *exc, const char *why, ...)
{
  va_list ap;

  va_start(ap, why);
  report_lost_exception_v(exc, why, ap);
  va_end(ap);
}

void stash_exception(struct excinfo *exc, const char *why, ...)
{
  va_list ap;
  struct excinfo stash;

  if (!exc->ty)
    STASH_EXCINFO(exc);
  else {
    va_start(ap, why);
    STASH_EXCINFO(&stash);
    report_lost_exception_v(&stash, why, ap);
    va_end(ap);
  }
}

void restore_exception(struct excinfo *exc, const char *why, ...)
{
  va_list ap;
  struct excinfo stash;

  if (!PyErr_Occurred())
    RESTORE_EXCINFO(exc);
  else {
    va_start(ap, why);
    STASH_EXCINFO(&stash);
    report_lost_exception_v(exc, why, ap);
    RESTORE_EXCINFO(&stash);
    va_end(ap);
  }
}

/*----- Generic dictionary methods ----------------------------------------*/

static PyTypeObject *itemiter_pytype, *valiter_pytype;

typedef struct iter_pyobj {
  PyObject_HEAD
  PyObject *map;
  PyObject *i;
} iter_pyobj;
#define ITER_MAP(o) (((iter_pyobj *)(o))->map)
#define ITER_I(o) (((iter_pyobj *)(o))->i)

static void iter_pydealloc(PyObject *me)
  { Py_DECREF(ITER_MAP(me)); Py_DECREF(ITER_I(me)); FREEOBJ(me); }

static PyObject *itemiter_pynext(PyObject *me)
{
  PyObject *k = 0, *v = 0, *rc = 0;

  if ((k = PyIter_Next(ITER_I(me))) != 0 &&
      (v = PyObject_GetItem(ITER_MAP(me), k)) != 0)
    rc = Py_BuildValue("(OO)", k, v);
  Py_XDECREF(k); Py_XDECREF(v);
  return (rc);
}

static PyTypeObject itemiter_pytype_skel = {
  PyObject_HEAD_INIT(0) 0,		/* Header */
  "ItemIter",				/* @tp_name@ */
  sizeof(iter_pyobj),			/* @tp_basicsize@ */
  0,					/* @tp_itemsize@ */

  iter_pydealloc,			/* @tp_dealloc@ */
  0,					/* @tp_print@ */
  0,					/* @tp_getattr@ */
  0,					/* @tp_setattr@ */
  0,					/* @tp_compare@ */
  0,					/* @tp_repr@ */
  0,					/* @tp_as_number@ */
  0,					/* @tp_as_sequence@ */
  0,					/* @tp_as_mapping@ */
  0,					/* @tp_hash@ */
  0,					/* @tp_call@ */
  0,					/* @tp_str@ */
  0,					/* @tp_getattro@ */
  0,					/* @tp_setattro@ */
  0,					/* @tp_as_buffer@ */
  Py_TPFLAGS_DEFAULT |			/* @tp_flags@ */
    Py_TPFLAGS_BASETYPE,

  /* @tp_doc@ */
  "Iterates over the items of a mapping.",

  0,					/* @tp_traverse@ */
  0,					/* @tp_clear@ */
  0,					/* @tp_richcompare@ */
  0,					/* @tp_weaklistoffset@ */
  PyObject_SelfIter,			/* @tp_iter@ */
  itemiter_pynext,			/* @tp_iternext@ */
  0,					/* @tp_methods@ */
  0,					/* @tp_members@ */
  0,					/* @tp_getset@ */
  0,					/* @tp_base@ */
  0,					/* @tp_dict@ */
  0,					/* @tp_descr_get@ */
  0,					/* @tp_descr_set@ */
  0,					/* @tp_dictoffset@ */
  0,					/* @tp_init@ */
  PyType_GenericAlloc,			/* @tp_alloc@ */
  abstract_pynew,			/* @tp_new@ */
  0,					/* @tp_free@ */
  0					/* @tp_is_gc@ */
};

static PyObject *valiter_pynext(PyObject *me)
{
  PyObject *k = 0, *rc = 0;

  if ((k = PyIter_Next(ITER_I(me))) != 0)
    rc = PyObject_GetItem(ITER_MAP(me), k);
  Py_XDECREF(k);
  return (rc);
}

static PyTypeObject valiter_pytype_skel = {
  PyObject_HEAD_INIT(0) 0,		/* Header */
  "ValueIter",				/* @tp_name@ */
  sizeof(iter_pyobj),			/* @tp_basicsize@ */
  0,					/* @tp_itemsize@ */

  iter_pydealloc,			/* @tp_dealloc@ */
  0,					/* @tp_print@ */
  0,					/* @tp_getattr@ */
  0,					/* @tp_setattr@ */
  0,					/* @tp_compare@ */
  0,					/* @tp_repr@ */
  0,					/* @tp_as_number@ */
  0,					/* @tp_as_sequence@ */
  0,					/* @tp_as_mapping@ */
  0,					/* @tp_hash@ */
  0,					/* @tp_call@ */
  0,					/* @tp_str@ */
  0,					/* @tp_getattro@ */
  0,					/* @tp_setattro@ */
  0,					/* @tp_as_buffer@ */
  Py_TPFLAGS_DEFAULT |			/* @tp_flags@ */
    Py_TPFLAGS_BASETYPE,

  /* @tp_doc@ */
  "Iterates over the values of a mapping.",

  0,					/* @tp_traverse@ */
  0,					/* @tp_clear@ */
  0,					/* @tp_richcompare@ */
  0,					/* @tp_weaklistoffset@ */
  PyObject_SelfIter,			/* @tp_iter@ */
  valiter_pynext,			/* @tp_iternext@ */
  0,					/* @tp_methods@ */
  0,					/* @tp_members@ */
  0,					/* @tp_getset@ */
  0,					/* @tp_base@ */
  0,					/* @tp_dict@ */
  0,					/* @tp_descr_get@ */
  0,					/* @tp_descr_set@ */
  0,					/* @tp_dictoffset@ */
  0,					/* @tp_init@ */
  PyType_GenericAlloc,			/* @tp_alloc@ */
  abstract_pynew,			/* @tp_new@ */
  0,					/* @tp_free@ */
  0					/* @tp_is_gc@ */
};

PySequenceMethods gmap_pysequence = {
  0,					/* @sq_length@ */
  0,					/* @sq_concat@ */
  0,					/* @sq_repeat@ */
  0,					/* @sq_item@ */
  0,					/* @sq_slice@ */
  0,					/* @sq_ass_item@ */
  0,					/* @sq_ass_slice@ */
  PyMapping_HasKey,			/* @sq_contains@ */
  0,					/* @sq_inplace_concat@ */
  0					/* @sq_inplace_repeat@ */
};

Py_ssize_t gmap_pysize(PyObject *me)
{
  PyObject *i = 0, *x = 0;
  Py_ssize_t rc = -1, n = 0;

  if ((i = PyObject_GetIter(me)) == 0) goto done;
  while ((x = PyIter_Next(i)) != 0) { n++; Py_DECREF(x); x = 0; }
  if (PyErr_Occurred()) goto done;
  rc = n;
done:
  Py_XDECREF(i); Py_XDECREF(x);
  return (rc);
}

PyObject *gmapmeth_has_key(PyObject *me, PyObject *arg)
{
  PyObject *k;
  if (!PyArg_ParseTuple(arg, "O:has_key", &k)) return (0);
  return (getbool(PyMapping_HasKey(me, k)));
}

PyObject *gmapmeth_keys(PyObject *me, PyObject *arg)
{
  PyObject *l = 0, *i = 0, *k, *rc = 0;
  int err;

  if (!PyArg_ParseTuple(arg, ":keys") ||
      (l = PyList_New(0)) == 0 ||
      (i = PyObject_GetIter(me)) == 0)
    goto done;
  while ((k = PyIter_Next(i)) != 0)
    { err = PyList_Append(l, k); Py_DECREF(k); if (err) goto done; }
  if (PyErr_Occurred()) goto done;
  rc = l; l = 0;
done:
  Py_XDECREF(l); Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_values(PyObject *me, PyObject *arg)
{
  PyObject *l = 0, *i = 0, *k, *v, *rc = 0;
  int err = 0;

  if (!PyArg_ParseTuple(arg, ":values") ||
      (l = PyList_New(0)) == 0 ||
      (i = PyObject_GetIter(me)) == 0)
    goto done;
  while ((k = PyIter_Next(i)) != 0) {
    if ((v = PyObject_GetItem(me, k)) == 0 ||
	PyList_Append(l, v))
      err = -1;
    Py_DECREF(k); Py_XDECREF(v);
    if (err) goto done;
  }
  if (PyErr_Occurred()) goto done;
  rc = l; l = 0;
done:
  Py_XDECREF(l); Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_items(PyObject *me, PyObject *arg)
{
  PyObject *l = 0, *i = 0, *k, *v, *z, *rc = 0;
  int err = 0;

  if (!PyArg_ParseTuple(arg, ":items") ||
      (l = PyList_New(0)) == 0 ||
      (i = PyObject_GetIter(me)) == 0)
    goto done;
  while ((k = PyIter_Next(i)) != 0) {
    z = 0;
    if ((v = PyObject_GetItem(me, k)) == 0 ||
	(z = Py_BuildValue("(OO)", k, v)) == 0 ||
	PyList_Append(l, z))
      err = -1;
    Py_DECREF(k); Py_XDECREF(v); Py_XDECREF(z);
    if (err) goto done;
  }
  if (PyErr_Occurred()) goto done;
  rc = l; l = 0;
done:
  Py_XDECREF(l); Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_iterkeys(PyObject *me, PyObject *arg)
{
  if (!PyArg_ParseTuple(arg, ":iterkeys")) return (0);
  return (PyObject_GetIter(me));
}

PyObject *gmapmeth_itervalues(PyObject *me, PyObject *arg)
{
  PyObject *i;
  iter_pyobj *ii;

  if (!PyArg_ParseTuple(arg, ":itervalues") ||
      (i = PyObject_GetIter(me)) == 0)
    return (0);
  ii = PyObject_NEW(iter_pyobj, valiter_pytype);
  ii->map = me; Py_INCREF(me);
  ii->i = i;
  return ((PyObject *)ii);
}

PyObject *gmapmeth_iteritems(PyObject *me, PyObject *arg)
{
  PyObject *i;
  iter_pyobj *ii;

  if (!PyArg_ParseTuple(arg, ":iteritems") ||
      (i = PyObject_GetIter(me)) == 0)
    return (0);
  ii = PyObject_NEW(iter_pyobj, itemiter_pytype);
  ii->map = me; Py_INCREF(me);
  ii->i = i;
  return ((PyObject *)ii);
}

PyObject *gmapmeth_clear(PyObject *me, PyObject *arg)
{
  PyObject *i = 0, *k = 0, *rc = 0;

  if (!PyArg_ParseTuple(arg, ":clear") ||
      (i = PyObject_GetIter(me)) == 0)
    goto end;
  while ((k = PyIter_Next(i)) != 0) {
    PyObject_DelItem(me, k);
    Py_DECREF(k);
  }
  if (PyErr_Occurred()) goto end;
  rc = me; Py_INCREF(me);
end:
  Py_XDECREF(i);
  return (rc);
}

static const char *const def_kwlist[] = { "key", "default", 0 };

PyObject *gmapmeth_get(PyObject *me, PyObject *arg, PyObject *kw)
{
  PyObject *k, *def = Py_None, *v;

  if (!PyArg_ParseTupleAndKeywords(arg, kw, "O|O:get",
				   (/*unconst*/ char **)def_kwlist,
				   &k, &def))
    return (0);
  if ((v = PyObject_GetItem(me, k)) != 0) return (v);
  PyErr_Clear();
  RETURN_OBJ(def);
}

PyObject *gmapmeth_setdefault(PyObject *me, PyObject *arg, PyObject *kw)
{
  PyObject *k, *def = Py_None, *v;

  if (!PyArg_ParseTupleAndKeywords(arg, kw, "O|O:setdefault",
				   (/*unconst*/ char **)def_kwlist,
				   &k, &def))
    return (0);
  if ((v = PyObject_GetItem(me, k)) != 0) return (v);
  PyErr_Clear();
  if (PyObject_SetItem(me, k, def)) return (0);
  RETURN_OBJ(def);
}

PyObject *gmapmeth_pop(PyObject *me, PyObject *arg, PyObject *kw)
{
  PyObject *k, *def = 0, *v;

  if (!PyArg_ParseTupleAndKeywords(arg, kw, "O|O:pop",
				   (/*unconst*/ char **)def_kwlist,
				   &k, &def))
    return (0);
  if ((v = PyObject_GetItem(me, k)) != 0) {
    PyObject_DelItem(me, k);
    return (v);
  } else if (def) {
    PyErr_Clear();
    RETURN_OBJ(def);
  } else
    return (0);
}

PyObject *gmapmeth_update(PyObject *me, PyObject *arg)
{
  PyObject *map, *i = 0, *k, *v, *rc = 0;
  int err = 0;

  if (!PyArg_ParseTuple(arg, "O:update", &map) ||
      (i = PyObject_GetIter(map)) == 0)
    goto end;
  while ((k = PyIter_Next(i)) != 0) {
    if ((v = PyObject_GetItem(map, k)) == 0 ||
	PyObject_SetItem(me, k, v))
      err = -1;
    Py_DECREF(k); Py_XDECREF(v);
    if (err) goto end;
  }
  if (PyErr_Occurred()) goto end;
  rc = me; Py_INCREF(me);
end:
  Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_popitem(PyObject *me, PyObject *arg)
{
  PyObject *i = 0, *k = 0, *v = 0, *rc = 0;

  if (!PyArg_ParseTuple(arg, ":popitem") ||
      (i = PyObject_GetIter(me)) == 0)
    goto end;
  if ((k = PyIter_Next(i)) == 0) {
    if (!PyErr_Occurred()) VALERR("popitem(): mapping is empty");
    goto end;
  }
  if ((v = PyObject_GetItem(me, k)) == 0 ||
      PyObject_DelItem(me, k))
    goto end;
  rc = Py_BuildValue("(OO)", k, v);
end:
  Py_XDECREF(i); Py_XDECREF(k); Py_XDECREF(v);
  return (rc);
}

PyMethodDef gmap_pymethods[] = {
  GMAP_METHODS
  { 0 }
};

/*----- Initialization ----------------------------------------------------*/

static PyObject *meth__set_home_module(PyObject *me, PyObject *arg)
{
  PyObject *mod;

  if (!PyArg_ParseTuple(arg, "O!:_set_home_module", &PyModule_Type, &mod))
    return (0);
  Py_XDECREF(home_module); home_module = mod; Py_INCREF(home_module);
  RETURN_NONE;
}

static const PyMethodDef methods[] = {
#define METHNAME(func) meth_##func
  METH	(_set_home_module, "_set_home_module(MOD)")
#undef METHNAME
  { 0 }
};

void util_pyinit(void)
{
  INITTYPE(itemiter, root);
  INITTYPE(valiter, root);
  addmethods(methods);
}

void util_pyinsert(PyObject *mod)
{
  INSERT("ItemIter", itemiter_pytype);
  INSERT("ValueIter", valiter_pytype);
}

/*----- That's all, folks -------------------------------------------------*/
